/*
   Silent clock driver using CNC Shield V4 driver board
     Can also use Silent Shield Clock Controller on Etsy

   Build a minimum hardware stepper motor clock driver using
     CNC Shield V4, TMC2208 stepper driver, and DS3231 RTC

   CNC Shield V4 is modified to fix jumper connections
     Jumpers are removed from lower right TMC2208 port
       All 3 pins hard wired to VIO (5V)
       TMC2208 motor supply voltage hard wired to 5V
     One motor header is hacked to connect DS3231 RTC
       Pin 1 connected to Vss
         (RTC NC and Vss pins shorted together)
       Pin 2 connected to SCL
       Pin 3 connected to SDA
       Pin 4 connected to VIO (5V)

   Use a DS3231 precision RTC as a reference time beat
     Accuracy 2PPM = about 1 minute per year drift
     Typical accuracy less than 1 second per month

   Plug TMC2208 into lower right port
     ENN  = D8 (PB0) Low = motors enabled
     DIR  = D4 (PD4)
     STEP = D7 (PD7)

   Jumpers in the 8 pin header set speed or set up debug modes
     CoolEN selects debug mode if inserted
     Resume, Hold, and Abort set speed to one of 8 speeds
       Parameters calculated any time jumpers are changed


   Download this sketch from https://www.stevesclocks.com/sp10
     Cut and paste the code into the Arduino IDE
     (Or change the file extension from .txt to .ino)


   Motor speed settings (subject to change)
     0:  0000  1.000 RPM  reserved
     1:  0001  4.630 RPM  SP10 with 200 step motors
     2:  0010  4.630 RPM  SP10 with 400 step motors
     3:  0011  3.600 RPM  SP6  with 200 step motors
     4:  0100  4.667 RPM  SP9  with 200 step motors
     5:  0101  6.944 RPM  SP11 with 200 step motors
     6:  0110  6.944 RPM  SP11 with 400 step motors
     7:  0111  20.00 RPM  fast debug mode

   Debug routines
     8:  1000  LED off     no test running
     9:  1001  LED on      simple hardware test
    10:  1010  LED on      simple hardware test
    11:  1011  Delay test  LED will blink at 1Hz (0.5s on, 0.5s off)
    12:  1100  LED on      simple hardware test
    13:  1101  RTC test    LED should blink at 0.5Hz if RTC is working
    14:  1110  Rotate CW   Rotate motor one direction at medium speed
    15:  1111  Rotate CCW  Rotate motor in opposite direction at high speed
    

   Revision History:
    05-Jan-23  1.00  Original release
    06-Jan-23  1.01  Switched from RTC.h to DS3231.h
    11-Jan-23  1.02  Updated debug monitor
    25-Jan-23  1.03  Added 4.667 speed ratio for SP9
    26-Jan-23  1.04  Added status LED monitor (on for "+")
    10-Feb-23  1.05  Swtched back to RTC.h (more stable)
    18-Feb-23  1.06  Moved motor delay calculations inside switch statement
    03-Mar-23  2.00  Moved I2C routines inline (no library installs needed)
    07-Mar-23  2.01  Added RTC timeout check
    25-Mar-23  2.02  Added debug modes
    25-Mar-23  2.03  Moved debug mode pin to D12
    08-Apr-23  2.04  Updated for SP10 and SP11
    13-Apr-23  2.05  Added modes for 400 step motors
    20-Apr-23  2.06  Reduced count on first loop to sync RTC faster
    21-Apr-23  2.07  Skipped one second for print delay at speed change
    16-Jul-23  2.08  Added comments for clock names at each speed change
*/

#define REVISION 2.08

#include <time.h>
#include <Wire.h>


//////////////////////////////////////
// Inline code to replace I2C_RTC.h //
// (truncated to minimal code size) //
//////////////////////////////////////
/* ------------------------------------------------------------
 * "THE BEERWARE LICENSE" (Revision 42):
 * <cvmanjoo@gmail.com> wrote this code. As long as you retain this
 * notice, you can do whatever you want with this stuff. If we
 * meet someday, and you think this stuff is worth it, you can
 * buy me a beer in return.
 * ------------------------------------------------------------
 * ------------------------------------------------------------
 * RTC.h - Library to set & get time from I2C RTCs for Arduino
 * Created by Manjunath CV. July 08, 2017, 02:18 AM
 * Released into the public domain.
 * -----------------------------------------------------------*/

#define DS3231_ADDR 0x68

class DS3231 {
  public:
    bool begin();
    uint8_t getSeconds();
};

///////////////////
// End I2C_RTC.h //
///////////////////


///////////////////////////////////////
// Inline code to replace DS3231.cpp //
// (truncated to minimal code size)  //
///////////////////////////////////////
/*
 * DS3231.cpp - Library to set & get time from RTC DS3231
 * Created by Manjunath CV. July 08, 2017, 02:18 AM
 * Released into the public domain.
 */

bool DS3231::begin() {
  Wire.begin(); // join i2c bus
  Wire.beginTransmission(0x68);
  return (Wire.endTransmission() == 0 ? true : false);
}

uint8_t DS3231::getSeconds() {
  uint8_t seconds;
  Wire.beginTransmission(DS3231_ADDR);
  Wire.write(0x00);  // Second Register
  Wire.endTransmission();
  Wire.requestFrom(DS3231_ADDR, 1);
  seconds = Wire.read();
  return (seconds);
}

////////////////////
// End DS3231.cpp //
////////////////////


///////////////////////////////////////////
// Start of code for silent clock driver //
///////////////////////////////////////////

static DS3231 RTC;        // RTC register definition

// define pins used
#define MOTOR_ENN_PIN       8
#define MOTOR_DIR_PIN       4
#define MOTOR_STEP_PIN      7
#define STATUS_LED_PIN     13
#define JUMPER_DEBUG_PIN   A3
#define JUMPER_SPEED_PIN_2 A2
#define JUMPER_SPEED_PIN_1 A1
#define JUMPER_SPEED_PIN_0 A0


// make most variables global (accessible during setup() and loop()
int motor_delay;            // delay value sent to motor
int min_motor_delay;        // short delay value
int max_motor_delay;        // long delay value
int motor_steps;            // number of steps per second
int extra_step_time;        // how often an extra step is needed
int extra_steps;            // how many extra steps get added
int extra_step_count = 0;   // counter for when to add extra steps
long seconds_counted = 0;   // total count of seconds elapsed
long seconds_moved = 0;     // total count of seconds motor has moved

int char_counter = 0;       // character display count (for CR/LF every minute)

int motor_speed_pins;       // current reading of 3 motor speed pins
int last_motor_speed = -1;  // previous reading of 3 motor speed pins
int rtc_seconds;            // actual seconds value returned by RTC
int last_seconds;           // previous seconds from RTC to compare against


void setup() {
  // put your setup code here, to run once:

  int i=0;                  // local counter for testing RTC

  pinMode(STATUS_LED_PIN,     OUTPUT);        // status LED
  pinMode(MOTOR_ENN_PIN,      OUTPUT);        // motor enable#
  pinMode(MOTOR_DIR_PIN,      OUTPUT);        // motor direction
  pinMode(MOTOR_STEP_PIN,     OUTPUT);        // motor step
  pinMode(JUMPER_SPEED_PIN_2, INPUT_PULLUP);  // input #2 to set motor speed
  pinMode(JUMPER_SPEED_PIN_1, INPUT_PULLUP);  // input #1 to set motor speed
  pinMode(JUMPER_SPEED_PIN_0, INPUT_PULLUP);  // input #0 to set motor speed
  pinMode(JUMPER_DEBUG_PIN,   INPUT_PULLUP);  // jumper for debug mode
  
  digitalWrite(STATUS_LED_PIN, LOW);   // turn off status LED

  // debug message at program start
  Serial.begin(9600);
  Serial.println("");
  Serial.println("");
  Serial.print("www.stevesclocks.com clock movement - Rev ");
  Serial.println(REVISION);

  // wait for seconds to change at beginning of algorithm
  //   (algorithm might hang if RTC is not present)
  Serial.println();
  Serial.print("Waiting for RTC...");
  RTC.begin();
  rtc_seconds = RTC.getSeconds(); // initialize rtc_seconds setting
  last_seconds = rtc_seconds;
  while ((rtc_seconds == last_seconds) && (i++ <= 30000)) {
    rtc_seconds = RTC.getSeconds();
  }
  // print "done" or error message if timeout
  if (i <= 30000) {
    Serial.print("Done. (");
    Serial.print(i);
    Serial.println(")");
  } else {
    Serial.println("RTC timeout, continuing without RTC");
  }
}


void loop() {
  // put your main code here, to run repeatedly:

  // local variables
  int step_count;             // counter for one second of movement
  int debug_mode_pins;        // jumper pin values during debug mode


  // check if debug jumper is installed
  while (digitalRead(JUMPER_DEBUG_PIN) == 0) {
    debug_mode_pins = 8
                    + 4 * !digitalRead(JUMPER_SPEED_PIN_2)
                    + 2 * !digitalRead(JUMPER_SPEED_PIN_1)
                    + 1 * !digitalRead(JUMPER_SPEED_PIN_0);
    switch (debug_mode_pins) {
      case 8: // turn off status LED if no other jumpers are installed
          digitalWrite(STATUS_LED_PIN, LOW);   // turn off status LED
          delay(10);
        break;
      case 9: // one jumper installed --> light LED for 100ms
      case 10:
      case 12:
          digitalWrite(STATUS_LED_PIN, HIGH);   // turn on status LED
          delay(10);
        break;
      case 11: // simple 1Hz LED blink test (500ms high, 500ms low)
          digitalWrite(STATUS_LED_PIN, HIGH);   // turn on status LED
          delay(500);
          digitalWrite(STATUS_LED_PIN, LOW);    // turn off status LED
          delay(500);
        break;
      case 13: // RTC test 0.5Hz LED toggle if successful
          rtc_seconds = RTC.getSeconds();       // read RTC
          digitalWrite(STATUS_LED_PIN, rtc_seconds & 0x01); // toggle LED
          delay(10);
        break;
      case 14: // rotate motor in one direction
          digitalWrite(MOTOR_DIR_PIN, LOW);
          // pulse motor one complete cycle (up and down)
          digitalWrite(MOTOR_STEP_PIN, HIGH);
          delayMicroseconds(2000);
          digitalWrite(MOTOR_STEP_PIN, LOW);
          delayMicroseconds(2000);
        break;
      default: // rotate motor in opposite direction
          digitalWrite(MOTOR_DIR_PIN, HIGH);
          // pulse motor one complete cycle (up and down)
          digitalWrite(MOTOR_STEP_PIN, HIGH);
          delayMicroseconds(1000);
          digitalWrite(MOTOR_STEP_PIN, LOW);
          delayMicroseconds(1000);
        break;
    }

    digitalWrite(STATUS_LED_PIN, LOW);  // turn off status LED before looping back
    digitalWrite(MOTOR_DIR_PIN, LOW);   // reset motor direction to default
  } // end debug test code



  // read motor speed pins and determine if they have changed
  motor_speed_pins = 4 * !digitalRead(JUMPER_SPEED_PIN_2)
                   + 2 * !digitalRead(JUMPER_SPEED_PIN_1)
                   + 1 * !digitalRead(JUMPER_SPEED_PIN_0);
  if (motor_speed_pins != last_motor_speed) {
    // determine new speed based on speed pins
    // min_motor_delay is approx  960000us / motor_steps / 2
    // max_motor_delay is approx 1010000us / motor_steps / 2
    // (actual values determined experimentally for 50% + to - ratio)
    switch (motor_speed_pins) {
      case 0: // 1.000 RPM ( slow test mode - reserved for future use)
          motor_steps = 53;     // 3200 / 60 = 53.3333
          extra_steps = 1;      // 1 extra step every 3 seconds
          extra_step_time = 3;
          min_motor_delay = 9000;
          max_motor_delay = 9500;
        break;
      case 1: // 4.630 RPM (SP10 with 200 step motors)
          motor_steps = 246;    // 3200 / (60/25*15/25*9) = 246.9135802
          extra_steps = 74;     // 913580246 extra steps every 999999999 seconds
          extra_step_time = 81; // (74 extra steps every 81 seconds)
          min_motor_delay = 1930;
          max_motor_delay = 2040;
        break;
      case 2: // 4.630 RPM (SP10 with 400 step motors)
          motor_steps = 492;    // twice as many steps as case 1
          extra_steps = 148;
          extra_step_time = 81;
          min_motor_delay =  960;
          max_motor_delay = 1015;
        break;
      case 3: // 3.600 RPM (SP6 with 200 step motors)
          motor_steps = 192;    // 3200 * 36 / 10 / 60 = 192
          extra_step_time = 0;  // no extra steps needed
          min_motor_delay = 2490;
          max_motor_delay = 2625;
        break;
      case 4: // 4.667:1 (SP9 with 200 step motors)
          motor_steps = 248;    // 3200 * 84 / 18 / 60 = 248.888
          extra_steps = 8;      // 8 extra steps every 9 seconds
          extra_step_time = 9;
          min_motor_delay = 1920;
          max_motor_delay = 2030;
        break;
      case 5: // 6.944 RPM (SP11 with 200 step motors)
          motor_steps = 370;    // 3200 / (60/25*15/25*6) = 370.37037037
          extra_steps = 10;     // 370 extra steps every 999 seconds (10 per 27)
          extra_step_time = 27;
          min_motor_delay = 1280;
          max_motor_delay = 1360;
        break;
      case 6: // 6.944 RPM (SP11 with 400 step motors)
          motor_steps = 740;    // twice as many steps as case 5
          extra_steps = 20;
          extra_step_time = 27;
          min_motor_delay = 630;
          max_motor_delay = 670;
        break;
      default: // 20.000 RPM (fast test mode - reserved for future use)
          motor_steps = 1066;   // 3200 / 60 * 20 = 1066.667
          extra_steps = 2;      // 2 extra steps every 3 seconds
          extra_step_time = 3;
          min_motor_delay = 430;
          max_motor_delay = 465;
        break;
    } // end switch()

    // Display new speed settings when they change
    Serial.println("");
    // Add a comment specific to each speed
    switch (motor_speed_pins) {
      case 0:  Serial.println("speed 0 (1.000 RPM slow test mode)"); break;
      case 1:  Serial.println("speed 1 (SP10 with 200 step motors)"); break;
      case 2:  Serial.println("speed 2 (SP10 with 400 step motors)"); break;
      case 3:  Serial.println("speed 3 (SP6 with 200 step motors)"); break;
      case 4:  Serial.println("speed 4 (SP9 with 200 step motors)"); break;
      case 5:  Serial.println("speed 5 (SP11 with 200 step motors)"); break;
      case 6:  Serial.println("speed 6 (SP11 with 400 step motors)"); break;
      default: Serial.println("speed 7 (20.000 RPM fast debug mode)"); break;
    } // end switch()
    Serial.print("steps ");
    Serial.print(motor_steps);
    if (extra_step_time) { // only print this value if extra_step time is set
      Serial.print(" (plus ");
      Serial.print(extra_steps);
      Serial.print("/");
      Serial.print(extra_step_time);
      Serial.print(")");
    }
    Serial.println();
    Serial.print("min_delay ");
    Serial.println(min_motor_delay);
    Serial.print("max_delay ");
    Serial.println(max_motor_delay);

    Serial.println("");

    // start with minimum motor delay (may miss first time)
    motor_delay = min_motor_delay;

    seconds_counted = 0;
    seconds_moved = 1;  // start at 1 second to account for print delay
    Serial.print(" ");  // display one character since seconds_moved = 1
    char_counter = 1;

    last_motor_speed = motor_speed_pins; // save latest speed setting

  } // end if(motor_speed_pins != last_motor_speed){}


  // move motor for approximately one second
  for (step_count = 0; step_count < motor_steps; step_count++) {
    // pulse one complete cycle (up and down)
    digitalWrite(MOTOR_STEP_PIN, HIGH);
    delayMicroseconds(motor_delay);
    digitalWrite(MOTOR_STEP_PIN, LOW);
    delayMicroseconds(motor_delay);

    // check every few cycles to see if RTC has advanced
    //   also, check just before the end of the loop
    if (((step_count % 50) == 0) || (step_count == (motor_steps - 1))) {
      // check if seconds have advanced
      rtc_seconds = RTC.getSeconds();
      if (rtc_seconds != last_seconds) {
        seconds_counted++;          // RTC has changed, increment seconds_counted
        last_seconds = rtc_seconds; // reset last_seconds to current time
      }
    }
  }
  seconds_moved++;  // motor has moved motor_steps, increment seconds_moved

  // check if seconds counted is the same as seconds moved
  if (seconds_counted >= seconds_moved) {
    // delay is just right (or slightly long) if seconds match
    //   decrease motor_delay to speed up slightly
    motor_delay = min_motor_delay;
    Serial.print("-");
    digitalWrite(STATUS_LED_PIN, LOW);   // turn off status LED
  } else {
    // delay is too short if seconds do not match
    //   increase motor delay to max
    motor_delay = max_motor_delay;
    Serial.print("+");
    digitalWrite(STATUS_LED_PIN, HIGH);   // turn on status LED
  }

  // check if extra delay needs to be added
  if (extra_step_time) {
    if (++extra_step_count >= extra_step_time) {
      for (step_count = 0; step_count < extra_steps; step_count++) {
        // pulse one complete cycle (up and down)
        digitalWrite(MOTOR_STEP_PIN, HIGH);
        delayMicroseconds(motor_delay);
        digitalWrite(MOTOR_STEP_PIN, LOW);
        delayMicroseconds(motor_delay);
      }
      extra_step_count = 0;   // reset extra step counter
    }
  }

  // add a CR/LF once per minute
  if (++char_counter >= 60) {
    // extra debug info every minute
    Serial.print(" (");
    if ((seconds_moved / 3600) >= 24) { // print days
      Serial.print(seconds_moved / 3600 / 24);
      Serial.print("d");
    }
    if (seconds_moved >= 3600) { // print hours
      Serial.print((seconds_moved / 3600) % 24);
      Serial.print("h");
    }
    // print minutes
    Serial.print((seconds_moved / 60) % 60);
    Serial.print("m)");
    // print number of seconds counted
    Serial.print(" (");
    Serial.print(seconds_counted);
    Serial.println(")");

    char_counter = 0;
  }

} // end loop()
